from ocetrac.track import (
    _morphological_operations,
    _apply_mask,
    _label_either,
    _filter_area,
    _wrap,
    track,
)

import pytest
import xarray as xr
import numpy as np
import scipy.ndimage
from skimage.measure import regionprops 
from skimage.measure import label as label_np
import dask.array as dsa


def example_anomaly_data():
    x0 = [180, 225, 360, 80, 1, 360, 1]
    y0 = [0, 20, -50, 40, -50, 40, 40]
    sigma0 = [15, 25, 30, 10, 30, 15, 10]

    lon = np.arange(0, 360) + 0.5
    lat = np.arange(-90, 90) + 0.5
    x, y = np.meshgrid(lon, lat)

    def make_blobs(x0, y0, sigma0):
        blob = np.exp(-((x - x0)**2 + (y - y0)**2)/(2*sigma0**2))
        return blob

    features = {}
    for i in range(len(x0)):
        features[i] = make_blobs(x0[i], y0[i], sigma0[i])

    first_image = features[0]+features[1]+features[3] -.5

    da = xr.DataArray(first_image[np.newaxis,:,:], dims=['time','lat', 'lon'],
                      coords={'time':[1],'lat': lat, 'lon': lon})

    da_shift_01 = da.shift(lon=0, lat=-20, fill_value=-.5)
    da_shift_02 = da.shift(lon=0, lat=-40, fill_value=-.5)+(features[2]+features[4]+features[5]+features[6])
    da_shift_03 = da.shift(lon=0, lat=-40, fill_value=-.5)+(features[2]+features[5]+features[6])

    Anom = xr.concat((da,
                      da_shift_01,
                      da_shift_02,
                      da_shift_03,), dim='time')
    
    Anom['time'] = np.arange(1,5)
    
    mask = xr.DataArray(np.ones(Anom[0,:,:].shape), coords=Anom[0,:,:].coords)
    mask[60:90,120:190] = 0
    
    return Anom, mask


def test_morphological_operations():
    Anom, mask = example_anomaly_data()
    Anom_dask = Anom.chunk({'time': 1})
    binary_images = _morphological_operations(Anom_dask, radius=8)
    
    assert isinstance(binary_images.data, dsa.Array)
    
    ocetrac_guess = Anom.where(binary_images==True, drop=False, other=np.nan) 
    best_guess = Anom.where(Anom>0, drop=False, other=np.nan) 
    part = ocetrac_guess.isin(best_guess)
    whole = best_guess.isin(best_guess)

    assert part.sum().values/whole.sum().values*100 >= 80
    assert part.sum().values == 26122
    
    
def test_apply_mask():
    Anom, mask = example_anomaly_data()
    Anom_dask = Anom.chunk({'time': 1})
    binary_images = _morphological_operations(Anom_dask, radius=8)
    binary_images_with_mask = binary_images.where(mask==1, drop=False, other=0)
    
    assert (binary_images_with_mask.where(mask==0, drop=True)==0).all()
    

def test_label_either():
    Anom, mask = example_anomaly_data()
    Anom_dask = Anom.chunk({'time': 1})
    binary_images = _morphological_operations(Anom_dask, radius=8)
    binary_images_with_mask = binary_images.where(mask==1, drop=False, other=0)

    labels, num = _label_either(binary_images_with_mask, return_num= True, connectivity=3)
    
    assert labels[2,:,:].max() == 6.
    assert all([i in labels[2,:,:] for i in range(0,6)])
    

def test_filter_area():
    Anom, mask = example_anomaly_data()
    Anom_dask = Anom.chunk({'time': 1})
    binary_images = _morphological_operations(Anom_dask, radius=8)
    binary_images_with_mask = binary_images.where(mask==1, drop=False, other=0)
    labels, num = _label_either(binary_images_with_mask, return_num= True, connectivity=3)
    
    min_size_quartile = .75
    area, min_area, labels_greater_minsize, labelprops = _filter_area(labels, min_size_quartile)
    
    assert min_area == 3413.0
    assert all([i in labels_greater_minsize for i in [1,4]])
    assert (area.label[area>=min_area].values == np.unique(labels_greater_minsize[labels_greater_minsize>0])).all()
    
    
def test_wrap():
    Anom, mask = example_anomaly_data()
    Anom_dask = Anom.chunk({'time': 1})
    binary_images = _morphological_operations(Anom_dask, radius=8)
    binary_images_with_mask = binary_images.where(mask==1, drop=False, other=0)
    labels, num = _label_either(binary_images_with_mask, return_num= True, connectivity=3)
    min_size_quartile = .75
    area, min_area, labels_greater_minsize, labelprops = _filter_area(labels, min_size_quartile)
    
    out_labels, N = _wrap(labels_greater_minsize)
    
    assert np.max(out_labels) == 2


@pytest.mark.parametrize("radius", [8, 10])
@pytest.mark.parametrize("area_quantile", [0.75, 0.80])
def test_track(radius, area_quantile):
    Anom, mask = example_anomaly_data()
    Anom_dask = Anom.chunk({'time': 1})
    new_labels = track(Anom, mask, radius=radius, area_quantile=area_quantile)
    
    assert (new_labels.percent_area_reject + new_labels.percent_area_accept) == 1.0

    props = regionprops(new_labels.astype('int'))
    labelprops = [p.label for p in props]
    labelprops = xr.DataArray(labelprops, dims=['label'], coords={'label': labelprops}) 
    area = xr.DataArray([p.area for p in props], dims=['label'], coords={'label': labelprops}) 
    
    assert area.min() >= new_labels.min_area
    


import xarray as xr
import numpy as np
import scipy.ndimage
from skimage.measure import regionprops 
from skimage.measure import label as label_np
import dask.array as dsa



def _morphological_operations(da, radius=8): 
    '''Converts xarray.DataArray to binary, defines structuring element, and performs morphological closing then opening.
    Parameters
    ----------
    da     : xarray.DataArray
            The data to label
    radius : int
            Length of grid spacing to define the radius of the structing element used in morphological closing and opening.
        
    '''
    
    # Convert images to binary. All positive values == 1, otherwise == 0
    bitmap_binary = da.where(da>0, drop=False, other=0)
    bitmap_binary = bitmap_binary.where(bitmap_binary==0, drop=False, other=1)
    
    # Define structuring element
    diameter = radius*2
    x = np.arange(-radius, radius+1)
    x, y = np.meshgrid(x, x)
    r = x**2+y**2 
    se = r<radius**2

    def binary_open_close(bitmap_binary):
        bitmap_binary_padded = np.pad(bitmap_binary,
                                      ((diameter, diameter), (diameter, diameter)),
                                      mode='wrap')
        s1 = scipy.ndimage.binary_closing(bitmap_binary_padded, se, iterations=1)
        s2 = scipy.ndimage.binary_opening(s1, se, iterations=1)
        unpadded= s2[diameter:-diameter, diameter:-diameter]
        return unpadded
    
    mo_binary = xr.apply_ufunc(binary_open_close, bitmap_binary,
                               input_core_dims=[['lat', 'lon']],
                               output_core_dims=[['lat', 'lon']],
                               output_dtypes=[bitmap_binary.dtype],
                               vectorize=True,
                               dask='parallelized')
    
    
    return mo_binary

def _apply_mask(mask, binary_images):
    binary_images_with_mask = binary_images.where(mask==1, drop=False, other=0)
    return binary_images_with_mask



def _label_either(data, **kwargs):
    if isinstance(data, dsa.Array):
        try:
            from dask_image.ndmeasure import label as label_dask
            def label_func(a, **kwargs):
                ids, num = label_dask(a, **kwargs)
                return ids
        except ImportError:
            raise ImportError(
                "Dask_image is required to use this function on Dask arrays. "
                "Either install dask_image or else call .load() on your data."
            )
    else:
        label_func = label_np
    return label_func(data, **kwargs)


def _filter_area(labels, min_size_quartile):
    '''calculatre area with regionprops'''
    
    props = regionprops(labels.astype('int'))
    labelprops = [p.label for p in props]
    labelprops = xr.DataArray(labelprops, dims=['label'], coords={'label': labelprops}) 
    area = xr.DataArray([p.area for p in props], dims=['label'], coords={'label': labelprops})  # Number of pixels of the region.
    min_area = np.percentile(area, min_size_quartile*100)
    keep_labels = labelprops.where(area>=min_area, drop=True)
    keep_where = np.isin(labels, keep_labels)
    labels_greater_minsize = np.where(keep_where==False, 0, labels)
    print('minimum area: ', min_area)  
    
    return area, min_area, labels_greater_minsize, labelprops


def _wrap(labels):
    ''' Impose periodic boundary and wrap labels'''
    first_column = labels[..., 0]
    last_column = labels[..., -1]
    
    unique_first = np.unique(first_column[first_column>0])
    
    # This loop iterates over the unique values in the first column, finds the location of those values in 
    # the first columnm and then uses that index to replace the values in the last column with the first column value
    for i in enumerate(unique_first):
        first = np.where(first_column == i[1])
        last = last_column[first[0], first[1]]
        bad_labels = np.unique(last[last>0])
        replace = np.isin(labels, bad_labels)
        labels[replace] = i[1]

    labels_wrapped = np.unique(labels, return_inverse=True)[1].reshape(labels.shape)
    
    # recalculate the total number of labels 
    N = np.max(labels_wrapped)

    return labels_wrapped, N


def track(da, mask, radius=8, area_quantile=0.75):
    '''Image labeling and tracking.
    
    Parameters
    ----------
    da : xarray.DataArray
        The data to label.
    
    radius : int
        size of the structuring element used in morphological opening and closing. Radius specified by the number of grid units.
        
    area_quantile : float
        quantile used to define the threshold of the smallest area object retained in tracking. Value should be between 0 and 1.
        
    mask : xarray.DataArray
        The mask of ponts to ignore. Must be binary where 1 = true point and 0 = background to be ignored. 
        
    Returns
    -------
    labels : xarray.DataArray
        Integer labels of the connected regions.
    '''
        
    # Convert data to binary, define structuring element, and perform morphological closing then opening
    binary_images = _morphological_operations(da, radius=8)
    
    # Apply mask
    binary_images_with_mask  = _apply_mask(mask, binary_images)
    
    # Label objects
    labels, num = _label_either(binary_images_with_mask, return_num= True, connectivity=3)
    
    # Filter area
    area, min_area, labels_greater_minsize, labelprops = _filter_area(labels, area_quantile)
    
    # Wrap labels
    labels_wrapped, N = _wrap(labels_greater_minsize)
    
    # Final labels to DataArray
    new_labels = xr.DataArray(labels_wrapped, dims=da.dims, coords=da.coords)   
    
    
    ## Metadata
    
    # Calculate Percent of total object area retained after size filtering
    sum_tot_area = int(np.sum(area.values))
    
    reject_area = area.where(area<=min_area, drop=True)
    sum_reject_area = int(np.sum(reject_area.values))
    percent_area_reject = (sum_reject_area/sum_tot_area)
    
    accept_area = area.where(area>min_area, drop=True)
    sum_accept_area = int(np.sum(accept_area.values))
    percent_area_accept = (sum_accept_area/sum_tot_area)

    new_labels = new_labels.rename('labels')
    new_labels.attrs['min_area'] = min_area
    new_labels.attrs['percent_area_reject'] = percent_area_reject
    new_labels.attrs['percent_area_accept'] = percent_area_accept
    
    print('inital objects identified \t', int(labels.max()))
    print('final objects tracked \t', int(N))
    
    return new_labels

